from plone.namedfile.file import NamedImage
from plone.namedfile.tests import getFile
from plone.namedfile.utils import get_contenttype
from plone.namedfile.utils import getImageInfo
from plone.namedfile.utils import safe_basename

import unittest


class TestUtils(unittest.TestCase):

    def test_get_contenttype(self):
        self.assertEqual(
            get_contenttype(
                NamedImage(
                    getFile("image.gif"),
                    contentType="image/gif",
                )
            ),
            "image/gif",
        )
        self.assertEqual(
            get_contenttype(
                NamedImage(
                    getFile("image.gif"),
                    filename="image.gif",
                )
            ),
            "image/gif",
        )
        self.assertEqual(
            get_contenttype(
                NamedImage(
                    getFile("image.tif"),
                    filename="image.tif",
                )
            ),
            "image/tiff",
        )
        self.assertEqual(
            get_contenttype(
                NamedImage(
                    getFile("notimage.doc"),
                    filename="notimage.doc",
                )
            ),
            "application/msword",
        )

        # Filename only detection of a non-IANA registered type.
        self.assertEqual(
            get_contenttype(filename="image.webp"),
            "image/webp",
        )

        # Filename only detection of a non-IANA registered type.
        self.assertEqual(
            get_contenttype(filename="song.midi"),
            "audio/midi",
        )

        # Detection of a surely not registered type.
        self.assertEqual(
            get_contenttype(filename="nothing.plonenamedfile"),
            "application/octet-stream",
        )

    def test_get_image_info(self):

        # WEBP WP8
        self.assertEqual(
            getImageInfo(getFile("image_lossy.webp")),
            ("image/webp", 500, 200),
        )

        # WEBP WP8L
        self.assertEqual(
            getImageInfo(getFile("image_loseless.webp")),
            ("image/webp", 200, 200),
        )

        # PNG
        self.assertEqual(
            getImageInfo(getFile("image.png")),
            ("image/png", 200, 200),
        )

        # BMP3
        self.assertEqual(
            getImageInfo(getFile("image.bmp")),
            ("image/x-ms-bmp", 200, 200),
        )

    def test_safe_basename_none(self):
        """filename=None is valid per INamed schema (required=False, default=None).
        Regression test for https://github.com/plone/Products.CMFPlone/issues/2633
        """
        self.assertIsNone(safe_basename(None))

    def test_safe_basename_unix(self):
        self.assertEqual(safe_basename("/path/to/file.txt"), "file.txt")

    def test_safe_basename_windows(self):
        self.assertEqual(safe_basename("C:\\path\\to\\file.txt"), "file.txt")

    def test_safe_basename_plain(self):
        self.assertEqual(safe_basename("file.txt"), "file.txt")



from OFS.SimpleItem import SimpleItem
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.base.utils import safe_text
from plone.namedfile import field
from plone.namedfile import file
from plone.namedfile.interfaces import IAvailableSizes
from plone.namedfile.interfaces import IImageScaleTraversable
from plone.namedfile.testing import PLONE_NAMEDFILE_FUNCTIONAL_TESTING
from plone.namedfile.tests import getFile
from plone.testing.zope import Browser
from zope.annotation import IAnnotations
from zope.annotation import IAttributeAnnotatable
from zope.component import getSiteManager
from zope.interface import implementer

import transaction
import unittest


class ISchema(IImageScaleTraversable):
    image = field.NamedImage()
    blob_image = field.NamedBlobImage()
    file = field.NamedFile()
    blob_file = field.NamedBlobFile()


@implementer(IAttributeAnnotatable, ISchema)
class DummyContent(SimpleItem):
    # Adapted from test_scaling_functional.py
    image = None
    blob_image = None
    file = None
    blob_file = None
    # modified = DateTime
    id = __name__ = "item"
    title = "foo"

    def Title(self):
        return self.title


def get_disposition_header(browser):
    # Could be CamelCase or all lowercase.
    name = "Content-Disposition"
    if name in browser.headers.keys():
        return browser.headers.get(name)
    name = name.lower()
    return browser.headers.get(name, None)


def custom_available_sizes():
    # Define available image scales.
    return {"custom": (10, 10)}


class TestAttackVectorNamedImage(unittest.TestCase):
    layer = PLONE_NAMEDFILE_FUNCTIONAL_TESTING
    field_class = file.NamedImage
    field_name = "image"

    def setUp(self):
        self.portal = self.layer["app"]
        item = DummyContent()
        self.layer["app"]._setOb("item", item)
        self.item = self.layer["app"].item
        sm = getSiteManager()
        sm.registerUtility(component=custom_available_sizes, provided=IAvailableSizes)

    def tearDown(self):
        sm = getSiteManager()
        sm.unregisterUtility(provided=IAvailableSizes)

    def get_admin_browser(self):
        browser = Browser(self.layer["app"])
        browser.handleErrors = False
        browser.addHeader(
            "Authorization",
            f"Basic {SITE_OWNER_NAME}:{SITE_OWNER_PASSWORD}",
        )
        return browser

    def get_anon_browser(self):
        browser = Browser(self.layer["app"])
        browser.handleErrors = False
        return browser

    def _named_file(self, name):
        data = getFile(name)
        return self.field_class(data, filename=safe_text(name))

    def assert_download_works(self, base_url):
        browser = self.get_anon_browser()
        browser.open(base_url + f"/@@download/{self.field_name}")
        header = get_disposition_header(browser)
        self.assertIsNotNone(header)
        self.assertIn("attachment", header)
        self.assertIn("filename", header)

    def assert_display_inline_works(self, base_url):
        # Test that displaying this file inline works.
        browser = self.get_anon_browser()
        browser.open(base_url + f"/@@display-file/{self.field_name}")
        self.assertIsNone(get_disposition_header(browser))

    def assert_display_inline_is_download(self, base_url):
        # Test that displaying this file inline turns into a download.
        browser = self.get_anon_browser()
        browser.open(base_url + f"/@@display-file/{self.field_name}")
        header = get_disposition_header(browser)
        self.assertIsNotNone(header)
        self.assertIn("attachment", header)
        self.assertIn("filename", header)

    def assert_scale_view_works(self, base_url):
        # Test that accessing a scale view shows the image inline.
        browser = self.get_anon_browser()
        browser.open(base_url + f"/@@images/{self.field_name}")
        self.assertIsNone(get_disposition_header(browser))

        # Note: the 'custom' scale is defined in an adapter above.
        browser.open(base_url + f"/@@images/{self.field_name}/custom")
        self.assertIsNone(get_disposition_header(browser))

        unique_scale_id = list(IAnnotations(self.item)["plone.scale"].keys())[0]
        browser.open(base_url + f"/@@images/{unique_scale_id}")
        self.assertIsNone(get_disposition_header(browser))

    def assert_scale_view_is_download(self, base_url):
        # Test that accessing a scale view turns into a download.
        browser = self.get_anon_browser()
        browser.open(base_url + f"/@@images/{self.field_name}")
        header = get_disposition_header(browser)
        self.assertIsNotNone(header)
        self.assertIn("attachment", header)
        self.assertIn("filename", header)

        browser.open(base_url + f"/@@images/{self.field_name}/custom")
        header = get_disposition_header(browser)
        self.assertIsNotNone(header)
        self.assertIn("attachment", header)
        self.assertIn("filename", header)

        unique_scale_id = list(IAnnotations(self.item)["plone.scale"].keys())[0]
        browser.open(base_url + f"/@@images/{unique_scale_id}")
        header = get_disposition_header(browser)
        self.assertIsNotNone(header)
        self.assertIn("attachment", header)
        self.assertIn("filename", header)

    def test_png_image(self):
        setattr(self.item, self.field_name, self._named_file("image.png"))
        transaction.commit()
        base_url = self.item.absolute_url()
        self.assert_download_works(base_url)
        self.assert_display_inline_works(base_url)
        if self.field_name == "image":
            self.assert_scale_view_works(base_url)

    def test_svg_image(self):
        setattr(self.item, self.field_name, self._named_file("image.svg"))
        transaction.commit()
        base_url = self.item.absolute_url()
        self.assert_download_works(base_url)
        self.assert_display_inline_is_download(base_url)
        if self.field_name == "image":
            self.assert_scale_view_is_download(base_url)

    def test_filename_none(self):
        # A 'None' filename probably does not happen during normal upload,
        # but if an attacker manages this, even @@download would show inline.
        # We prevent this.
        data = self._named_file("image.svg")
        data.filename = None
        setattr(self.item, self.field_name, data)
        transaction.commit()
        base_url = self.item.absolute_url()
        self.assert_download_works(base_url)
        self.assert_display_inline_is_download(base_url)
        if self.field_name == "image":
            self.assert_scale_view_is_download(base_url)

    def test_filename_empty(self):
        # An empty filename is probably no problem, but let's check.
        data = self._named_file("image.svg")
        data.filename = ""
        setattr(self.item, self.field_name, self._named_file("image.svg"))
        transaction.commit()
        base_url = self.item.absolute_url()
        self.assert_download_works(base_url)
        self.assert_display_inline_is_download(base_url)
        if self.field_name == "image":
            self.assert_scale_view_is_download(base_url)


class TestAttackVectorNamedBlobImage(TestAttackVectorNamedImage):
    field_class = file.NamedBlobImage


class TestAttackVectorNamedFile(TestAttackVectorNamedImage):
    field_class = file.NamedFile
    field_name = "file"

    def test_html_file(self):
        data = self.field_class(
            "<h1>Attacker</h1>", filename=safe_text("attacker.html")
        )
        setattr(self.item, self.field_name, data)
        transaction.commit()
        base_url = self.item.absolute_url()
        self.assert_download_works(base_url)
        self.assert_display_inline_is_download(base_url)

    def test_pdf(self):
        # By popular request we allow PDF.
        setattr(self.item, self.field_name, self._named_file("file.pdf"))
        transaction.commit()
        base_url = self.item.absolute_url()
        self.assert_download_works(base_url)
        self.assert_display_inline_works(base_url)


class TestAttackVectorNamedBlobFile(TestAttackVectorNamedFile):
    field_class = file.NamedBlobFile



# This file was borrowed from z3c.blobfile and is licensed under the terms of
# the ZPL.

##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from OFS.Image import Pdata
from plone.namedfile import field
from plone.namedfile.file import FileChunk
from plone.namedfile.file import NamedBlobImage
from plone.namedfile.testing import PLONE_NAMEDFILE_FUNCTIONAL_TESTING
from plone.namedfile.tests import getFile
from unittest.mock import patch
from ZODB.blob import Blob
from ZODB.blob import BlobFile

import io
import os
import piexif
import PIL
import tempfile
import unittest


class TestStorable(unittest.TestCase):

    layer = PLONE_NAMEDFILE_FUNCTIONAL_TESTING

    def test_pdata_storable(self):
        pdata = Pdata(getFile("image.gif"))
        fi = NamedBlobImage(pdata, filename="image.gif")
        self.assertEqual(303, fi.getSize())

    def test_str_storable(self):
        fi = NamedBlobImage(getFile("image.gif"), filename="image.gif")
        self.assertEqual(303, fi.getSize())

    def test_filechunk_storable(self):
        fi = NamedBlobImage(FileChunk(getFile("image.gif")), filename="image.gif")
        self.assertEqual(303, fi.getSize())

    def test_opened_file_storable(self):
        data = getFile("image.gif")
        f = tempfile.NamedTemporaryFile(delete=False)
        try:
            path = f.name
            f.write(data)
            f.close()
            with open(path, "rb") as f:
                fi = NamedBlobImage(f, filename="image.gif")
        finally:
            if os.path.exists(path):
                os.remove(path)
        self.assertEqual(303, fi.getSize())

    def test_upload_no_read(self):
        # ensure we don't read the whole file into memory

        old_open = Blob.open
        blob_read = 0
        blob_write = 0
        old_read = BlobFile.read
        read_bytes = 0

        def count_open(self, mode="r"):
            nonlocal blob_read, blob_write
            blob_read += 1 if "r" in mode else 0
            blob_write += 1 if "w" in mode else 0
            return old_open(self, mode)

        def count_reads(self, size=-1):
            nonlocal read_bytes
            res = old_read(self, size)
            read_bytes += len(res)
            return res

        with (
            patch.object(Blob, "open", count_open),
            patch.object(BlobFile, "read", count_reads),
        ):
            data = getFile("image.jpg")
            f = tempfile.NamedTemporaryFile(delete=False)
            try:
                path = f.name
                f.write(data)
                f.close()
                with open(path, "rb") as f:
                    fi = NamedBlobImage(f, filename="image.gif")
            finally:
                if os.path.exists(path):
                    os.remove(path)
            self.assertEqual(3641, fi.getSize())
            self.assertIn("Exif", fi.exif)
            self.assertEqual(500, fi._width)
            self.assertEqual(200, fi._height)
            self.assertLess(
                read_bytes,
                fi.getSize(),
                "Images should not need to read all data to get exif, dimensions",
            )
            self.assertEqual(
                blob_read, 3, "blob opening for getsize, get_exif and getImageInfo only"
            )
            self.assertEqual(
                blob_write,
                1,
                "Slow write to blob instead of os rename. Should be only 1 on __init__ to create empty file",
            )

            # Test validation which should also not read all the data.
            blob_read = read_bytes = 0

            blob_field = field.NamedBlobImage()
            blob_field.validate(fi)
            blob_field = field.NamedBlobFile()  # Image is subclass of file
            blob_field.validate(fi)

            self.assertEqual(
                blob_read, 0, "Validation is reading the whole blob in memory"
            )
            self.assertEqual(
                read_bytes, 0, "Validation is reading the whole blob in memory"
            )

    def test_large_webp_storable(self):
        # ensure we don't read the whole file into memory

        old_open = Blob.open
        blob_read = 0
        blob_write = 0
        old_read = BlobFile.read
        read_bytes = 0

        def count_open(self, mode="r"):
            nonlocal blob_read, blob_write
            blob_read += 1 if "r" in mode else 0
            blob_write += 1 if "w" in mode else 0
            return old_open(self, mode)

        def count_reads(self, size=-1):
            nonlocal read_bytes
            res = old_read(self, size)
            read_bytes += len(res)
            return res

        with (
            patch.object(Blob, "open", count_open),
            patch.object(BlobFile, "read", count_reads),
        ):
            fi = NamedBlobImage(getFile("900.webp"), filename="900.webp")
            self.assertEqual((900, 900), fi.getImageSize())
            self.assertLess(
                read_bytes,
                fi.getSize(),
                "Images should not need to read all data to get exif, dimensions",
            )
            self.assertEqual(
                blob_read, 3, "blob opening for getsize, get_exif and getImageInfo only"
            )

    def test_rotate(self):
        # Create a 200x200 white image
        img = PIL.Image.new("RGB", (100, 400), "white")
        # Set the top-left pixels to black
        img.putpixel((0, 0), (0, 0, 0))

        # 270 degree rotation
        # Create EXIF dict
        exif_dict = {"0th": {}, "Exif": {}, "GPS": {}, "1st": {}, "thumbnail": None}
        exif_dict["0th"][piexif.ImageIFD.Orientation] = 6
        exif_bytes = piexif.dump(exif_dict)
        # Save image as JPEG with EXIF data
        out = io.BytesIO()
        img.save(out, format="JPEG", exif=exif_bytes)
        fi_jpg = NamedBlobImage(out.getvalue(), filename="image.jpg")
        img_jpg = PIL.Image.open(io.BytesIO(fi_jpg.data))
        self.assertEqual(fi_jpg.getImageSize(), (400, 100))
        self.assertEqual(img_jpg.getpixel((0, 0)), (255, 255, 255))
        # The pixel is not exactly black (RGB 0,0,0) due to quantization errors and
        # compression artifacts introduced by the JPEG encoding process.
        self.assertEqual(img_jpg.getpixel((399, 0)), (10, 10, 10))
        self.assertEqual(img_jpg.getpixel((0, 99)), (255, 255, 255))
        self.assertEqual(img_jpg.getpixel((399, 99)), (255, 255, 255))

        # flip left to right
        # Create EXIF dict
        exif_dict = {"0th": {}, "Exif": {}, "GPS": {}, "1st": {}, "thumbnail": None}
        exif_dict["0th"][piexif.ImageIFD.Orientation] = 2
        exif_bytes = piexif.dump(exif_dict)
        # Save image as JPEG with EXIF data
        out = io.BytesIO()
        img.save(out, format="JPEG", exif=exif_bytes)
        fi_jpg = NamedBlobImage(out.getvalue(), filename="image.jpg")
        img_jpg = PIL.Image.open(io.BytesIO(fi_jpg.data))
        self.assertEqual(fi_jpg.getImageSize(), (100, 400))
        self.assertEqual(img_jpg.getpixel((0, 0)), (255, 255, 255))
        # The pixel is not exactly black (RGB 0,0,0) due to quantization errors and
        # compression artifacts introduced by the JPEG encoding process.
        self.assertEqual(img_jpg.getpixel((99, 0)), (8, 8, 8))
        self.assertEqual(img_jpg.getpixel((0, 399)), (255, 255, 255))
        self.assertEqual(img_jpg.getpixel((99, 399)), (255, 255, 255))



from pathlib import Path
from setuptools import setup

version = "8.0.0a3"

description = "File types and fields for images, files and blob files with filenames"
long_description = (
    f"{Path('README.rst').read_text()}\n"
    f"{Path('CHANGES.rst').read_text()}\n"
    f"{(Path('src') / 'plone' / 'namedfile' / 'usage.rst').read_text()}"
)

setup(
    name="plone.namedfile",
    version=version,
    description=description,
    long_description=long_description,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Plone",
        "Framework :: Plone :: 6.2",
        "Framework :: Plone :: Core",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: BSD License",
    ],
    keywords="plone named file image blob",
    author="Laurence Rowe, Martin Aspeli",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://pypi.org/project/plone.namedfile",
    license="BSD",
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.10",
    install_requires=[
        "BeautifulSoup4",
        "persistent",
        "piexif",
        "Pillow",
        "plone.app.uuid>=2.2.0",
        "plone.base",
        "plone.dexterity",
        "plone.memoize",
        "plone.protect",
        "plone.rfc822>=2.0.0",
        "plone.scale[storage]>=4.2.0",
        "plone.schemaeditor",
        "plone.supermodel",
        "Products.CMFCore",
        "z3c.caching",
        "zope.cachedescriptors",
        "zope.copy",
        "Zope",
    ],
    extras_require={
        "test": [
            "lxml",
            "plone.app.testing",
            "plone.testing",
        ],
    },
)
